SystemOrganization addCategory: #CenteredIntervalTree!

Object subclass: #CenteredIntervalTree
	instanceVariableNames: 'iMidLeft iMidRight leftSubTree rightSubTree xMid'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CenteredIntervalTree'!

!CenteredIntervalTree commentStamp: 'PaulDeBruicker 10/11/2011 12:50' prior: 0!
is much faster to find all intersecting timespan than selecting across a collection.  querying methods return a collection.  

From here: https://secure.wikimedia.org/wikipedia/en/wiki/Interval_tree

and: 

http://brettwalenz.org/2010/09/28/centered-interval-trees/!

!CenteredIntervalTree classSide methodsFor: 'instance creation' stamp: 'PaulDeBruicker 10/12/2011 06:55'!
from: aCollectionOfIntervals

	^ aCollectionOfIntervals isEmpty
		ifTrue: [ CenteredIntervalTreeSentinelNode default ]
		ifFalse: [ self new createFrom: aCollectionOfIntervals ]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/13/2011 10:54'!
checkForOverlap: aRange
	| col |
	col := OrderedCollection new.

	iMidLeft
		do: [ :each | (aRange includes:each)
				ifTrue: [ col add: each ]
				ifFalse: [ ^ col ]] .
			^col
! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 12:44'!
checkLeftSideFor: aPoint
	^ aPoint < xMid
		ifTrue: [ (self selectFrom: iMidLeft a: aPoint) , (leftSubTree intervalsThatInclude: aPoint) ]
		ifFalse: [ OrderedCollection new ]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/13/2011 09:05'!
checkLeftSideForRange: aRange
	^ aRange start < xMid
		ifTrue: [ (self selectFrom: iMidLeft a: aRange end) , (leftSubTree intervalsThatInclude: aRange) ]
		ifFalse: [ OrderedCollection new ]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 12:46'!
checkRightSideFor: aPoint
	^ aPoint > xMid
		ifTrue: [ (self selectFrom: iMidRight a: aPoint)  , (rightSubTree intervalsThatInclude: aPoint) ]
		ifFalse: [ OrderedCollection new ]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/13/2011 09:05'!
checkRightSideForRange: aRange
	^ aRange end > xMid
		ifTrue: [ (self selectFrom: iMidRight a: aRange start) , (rightSubTree intervalsThatInclude: aRange) ]
		ifFalse: [ OrderedCollection new ]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/12/2011 06:50'!
contents
	^ iMidLeft ,  leftSubTree contents , rightSubTree contents! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 10:48'!
createFrom: aCollectionOfIntervals 
 |iMid iLeft iRight|

	xMid:=(aCollectionOfIntervals collect:[:each | each end]) median.
	iMid:=aCollectionOfIntervals select:[:each | each start<= xMid and:[each end >= xMid]].
		iMidLeft:=iMid asSortedCollection:[:a :c | a start < c start].
		iMidRight:=iMid asSortedCollection:[:a :c | a end < c end].
		leftSubTree:=CenteredIntervalTree from:(aCollectionOfIntervals select: [:each | each end < xMid]).
		rightSubTree:=CenteredIntervalTree from:(aCollectionOfIntervals select: [:each | each start > xMid]).
	! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 15:59'!
end
	^self contents last end! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 10:51'!
initialize 
	iMidLeft:=OrderedCollection new.
	iMidRight:=OrderedCollection new.! !

!CenteredIntervalTree methodsFor: 'querying' stamp: 'PaulDeBruicker 10/11/2011 12:21'!
intervalsThatInclude: aPointOrRange
	^ (aPointOrRange respondsTo: #start)
		ifTrue: [ self intervalsThatIncludeRange: aPointOrRange ]
		ifFalse: [ self intervalsThatIncludePoint: aPointOrRange ]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/12/2011 06:52'!
intervalsThatIncludePoint: aPoint
	^ (self checkLeftSideFor: aPoint) , (self checkRightSideFor: aPoint)! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/13/2011 09:10'!
intervalsThatIncludeRange: aRange

	^
		(self checkLeftSideForRange:aRange) , (self checkRightSideForRange: aRange) , (self checkForOverlap: aRange)
		
! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 17:12'!
isEmpty
	^iMidLeft isEmpty and:[iMidRight isEmpty]! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/12/2011 06:48'!
isLeaf
	^ false! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 16:48'!
matches: anIntervalTree
	|matches|
	matches:=OrderedCollection new.
	anIntervalTree contents do:[:each |matches addAll: (self intervalsThatIncludeRange: each) ].
	^matches! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 17:38'!
medianValue
	^xMid! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 16:31'!
notEmpty
	^self isEmpty not! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/13/2011 09:40'!
selectFrom: aSortedCollection a: pointValue
^(aSortedCollection select:[:each | each includes: pointValue ]) asOrderedCollection ! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/12/2011 06:49'!
size
	^ self contents size! !

!CenteredIntervalTree methodsFor: 'initialization' stamp: 'PaulDeBruicker 10/11/2011 15:58'!
start
	^self contents first start! !

Object subclass: #CenteredIntervalTreeSentinelNode
	instanceVariableNames: 'emptyCollection'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'CenteredIntervalTree'!

!CenteredIntervalTreeSentinelNode classSide methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:54'!
default
	^Default ifNil:[Default:=self new]! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:50'!
contents
	^emptyCollection ! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:49'!
iMidLeft
	^emptyCollection ! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:49'!
iMidRight
	^emptyCollection ! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:49'!
initialize 
	emptyCollection :=OrderedCollection new.! !

!CenteredIntervalTreeSentinelNode methodsFor: 'testing' stamp: 'PaulDeBruicker 10/12/2011 06:56'!
intervalsThatInclude: aTimespan 
	^emptyCollection ! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:51'!
isEmpty 
	^true! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:47'!
isLeaf
	^true! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:51'!
notEmpty 
	^false! !

!CenteredIntervalTreeSentinelNode methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 06:48'!
size
	^0! !

TestCase subclass: #CenteredIntervalTreeTest
	instanceVariableNames: 'timespans pointInTime timespan intervalDatabase'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CenteredIntervalTree'!

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/13/2011 10:51'!
setUp

pointInTime:=DateAndTime year:2011 month:10 day: 13 hour: 12 minute: 0 second: 0 offset: Duration zero.
timespans:=(1 to: 100) collect:[:each |(Timespan starting: (pointInTime + (each-50) days) duration: (8 hours))] .

timespan:=Timespan starting: pointInTime + 15 minutes duration: 10 hours.

intervalDatabase:=(1to:10) collect:[:each | each to:(each * 20)].
! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/13/2011 10:54'!
testBasicQuery
	|cit int |
	cit:=CenteredIntervalTree from: timespans.
	self assert: (cit intervalsThatInclude: timespan) notEmpty.
	self assert: (cit intervalsThatInclude: pointInTime) notEmpty.
	cit:=CenteredIntervalTree from: intervalDatabase.
	int:=intervalDatabase atRandom.
	self assert: (cit intervalsThatInclude: int) notEmpty.
	self assert: (cit intervalsThatInclude: int start) notEmpty.
	self assert: (cit intervalsThatInclude: int end) notEmpty.! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/11/2011 11:04'!
testCreateFrom
	|cit |
	cit:=CenteredIntervalTree from: timespans.
	self assert: cit medianValue notNil.
	self assert: cit isEmpty equals: false.
! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/11/2011 11:58'!
testEnclosedTimespan
	|cit tspan |
	cit:=CenteredIntervalTree from: timespans.
	tspan:=Timespan starting: pointInTime + 1 hour  duration: 1 hour.
	self assert: (cit intervalsThatInclude: tspan) notEmpty. ! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/11/2011 12:23'!
testFaillingQuery
	|cit tspan |
	cit:=CenteredIntervalTree from: timespans.
	tspan:=Timespan starting: pointInTime + 5000 days duration: 1 hour.
	self assert: (cit intervalsThatInclude: tspan) isEmpty. "tspan is after the latest timespan in the tree"
	self assert: (cit intervalsThatInclude: pointInTime + 5000 days) isEmpty.
	tspan:=Timespan starting: pointInTime - 5000 days duration: 1 hour.	
	self assert: (cit intervalsThatInclude: tspan) isEmpty. "tspan is before  the earliest timespan in the tree"
	self assert: (cit intervalsThatInclude: pointInTime - 5000 days) isEmpty.
	! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/12/2011 08:43'!
testFindAllTimespans
	| cit col |
	cit:=CenteredIntervalTree from: timespans.
	col:=OrderedCollection new.
	timespans do:[:each | col addAll: (cit intervalsThatInclude: each) asSet].
	self assert: col size equals: timespans size.
! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/13/2011 10:58'!
testQuerySpeed
	| cit intersectionTime citTime |
	intersectionTime := [ 10 timesRepeat: [ timespans select: [ :each | (each intersection: timespan) notNil ] ] ]
		timeToRun.
	cit := CenteredIntervalTree from: timespans.
	citTime := [ 10 timesRepeat: [ cit intervalsThatInclude: timespan ] ] timeToRun.
	self assert: citTime < intersectionTime! !

!CenteredIntervalTreeTest methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 10/11/2011 12:01'!
testStartAndEndMatched
	|cit tspan |
	cit:=CenteredIntervalTree from: timespans.
	tspan:=Timespan starting: pointInTime duration: 1 hour.
	self assert: (cit intervalsThatInclude: tspan) notEmpty. "start matches"
		tspan:=Timespan starting: pointInTime duration: 8 hour.
	self assert: (cit intervalsThatInclude: tspan) notEmpty. "exact overlap" 
		tspan:=Timespan starting: pointInTime + 7 hours duration: 1 hour.
	self assert: (cit intervalsThatInclude: tspan) notEmpty. "end matches"
	! !

!Interval methodsFor: '*centeredintervaltree' stamp: 'PaulDeBruicker 10/13/2011 09:02'!
end
	^self last! !

!Interval methodsFor: '*centeredintervaltree' stamp: 'PaulDeBruicker 10/13/2011 10:57'!
includes: aNumberOrInterval
	^(aNumberOrInterval respondsTo: #end)
		ifTrue: [ start <= aNumberOrInterval end and: [ stop >= aNumberOrInterval start ] ]
		ifFalse: [ self rangeIncludes: aNumberOrInterval ]! !

!Interval methodsFor: '*centeredintervaltree' stamp: 'PaulDeBruicker 10/13/2011 09:03'!
start
	^self first! !
